// $Id: DocumentPanelEvent.java 15358 2012-03-29 19:06:36Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.event.*;
import java.awt.*;
import javax.swing.*;
import org.faceless.pdf2.*;

/**
 * Represents an event on a {@link DocumentPanel} indicating that document has changed
 * somehow. To capture these events, implement the {@link DocumentPanelListener} interface
 * and register via the {@link DocumentPanel#addDocumentPanelListener DocumentPanel.addDocumentPanelListener()} method.
 * Be sure to check the {@link #getType} method to see what sort of event it is - current values
 * include:
 * <table class="defntable">
 * <tr><th>created</th><td>Raised when the DocumentPanel is first created</code></td></tr>
 * <tr><th>activated</th><td>Raised when the DocumentPanel is activated with a valid PDF</code></td></tr>
 * <tr><th>deactivated</th><td>Raised when the DocumentPanel is deactivated</code></td></tr>
 * <tr><th>viewportChanged</th><td>Raised when the DocumentPanel has a new {@link DocumentViewport}  applied to it</code></td></tr>
 * <tr><th>loaded</th><td>Raised after a PDF is loaded</code></td></tr>
 * <tr><th>closing</th><td>Raised when the PDF is closing</code></td></tr>
 * <tr><th>redrawn</th><td>Raised when the PDF has been redrawn somehow</code></td></tr>
 * <tr><th>pageChanged</th><td>Raised when the DocumentPanel changes which page is being displayed</code></td></tr>
 * <tr><th>pagePositionChanged</th><td>Raised when the DocumentPanel changes which area of the page is being displayed</code></td></tr>
 * </table>
 * DocumentPanelEvent's are created with the static "create" methods, but unless you are
 * implementing your own Viewport, there is probably no need to call these methods.
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class DocumentPanelEvent {

    private final DocumentPanel panel;
    private final String type;

    /**
     * Create a "created" DocumentPanelEvent
     */
    public static DocumentPanelEvent createCreated(DocumentPanel panel) {
        return new DocumentPanelEvent("created", panel);
    }
    /**
     * Create an "activated" DocumentPanelEvent
     */
    public static DocumentPanelEvent createActivated(DocumentPanel panel) {
        return new DocumentPanelEvent("activated", panel);
    }
    /**
     * Create a "deactivated" DocumentPanelEvent
     */
    public static DocumentPanelEvent createDeactivated(DocumentPanel panel) {
        return new DocumentPanelEvent("deactivated", panel);
    }
    /**
     * Create a "viewportChanged" DocumentPanelEvent
     */
    public static DocumentPanelEvent createViewportChanged(DocumentPanel panel) {
        return new DocumentPanelEvent("viewportChanged", panel);
    }
    /**
     * Create a "pageChanged" DocumentPanelEvent
     */
    public static DocumentPanelEvent createPageChanged(DocumentPanel panel) {
        return new DocumentPanelEvent("pageChanged", panel);
    }
    /**
     * Create a "pagePositionChagned" DocumentPanelEvent
     */
    public static DocumentPanelEvent createPagePositionChanged(DocumentPanel panel) {
        return new DocumentPanelEvent("pagePositionChanged", panel);
    }
    /**
     * Create a "loaded" DocumentPanelEvent
     */
    public static DocumentPanelEvent createLoaded(DocumentPanel panel) {
        return new DocumentPanelEvent("loaded", panel);
    }
    /**
     * Create a "closing" DocumentPanelEvent
     */
    public static DocumentPanelEvent createClosing(DocumentPanel panel) {
        return new DocumentPanelEvent("closing", panel);
    }
    /**
     * Create a "redrawn" DocumentPanelEvent
     */
    public static DocumentPanelEvent createRedrawn(DocumentPanel panel) {
        return new DocumentPanelEvent("redrawn", panel);
    }

    private DocumentPanelEvent(String type, DocumentPanel panel) {
        this.type = type;
        this.panel = panel;
    }

    public String toString() {
        return "[D:"+type+"]";
    }

    /**
     * Get the DocumentPanel this event was raised on
     */
    public DocumentPanel getDocumentPanel() {
        return panel;
    }

    /**
     * Get the type of the DocumentPanelEvent
     */
    public String getType() {
        return type;
    }
}
