// $Id: PDFViewerApplet.java 15360 2012-03-29 19:12:25Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.util.List;
import java.net.URL;
import org.faceless.pdf2.PDF;
import org.faceless.pdf2.PropertyManager;
import org.faceless.pdf2.EncryptionHandler;
import org.faceless.pdf2.StandardEncryptionHandler;
import org.faceless.pdf2.viewer2.feature.About;
import org.faceless.pdf2.viewer2.feature.PDFImporter;

/**
 * An applet wrapping the {@link PDFViewer}. At it's simplest you can simply embed it in the
 * page using the <code>applet</code> tag, although since 2.8.5 this class can be usefully
 * controlled from JavaScript via LiveConnect. For instance, to instantiate an applet
 * that is controlled via JavaScript buttons instead of Java buttons:
 * <pre class="example" style="font-size:smaller">
 * &lt;html&gt;
 *  &lt;head&gt;
 *   &lt;style&gt;
 *    applet { width:100%; height:500px; display:block }
 *    form   { text-align:center  }
 *   &lt;/style&gt;
 *   &lt;script&gt;
 *    function documentUpdated(event) {
 *      if (event.getType()=="pageChanged") {
 *        document.forms[0].pagenumber.value = event.getDocumentPanel().getPageNumber()+1;
 *      }
 *    }
 *   &lt;/script&gt;
 *  &lt;/head&gt;
 *  &lt;body&gt;
 *   &lt;applet code="org.faceless.pdf2.viewer2.PDFViewerApplet" name="pdfapplet" archive="bfopdf.jar, bfopdf-cmap.jar"&gt;
 *    &lt;param name="pdf" value="test.pdf" /&gt;<sup><a href="#param1">1</a></sup>
 *    <i>&lt;!-- &lt;param name="password" value="secret" /&gt; --&gt;</i>
 *    &lt;param name="feature.MultiWindow" value="false" /&gt;<sup><a href="#param2">2</a></sup>
 *    &lt;param name="feature.Toolbars" value="false" /&gt;
 *    &lt;param name="feature.ThumbnailPanel.editable" value="false" /&gt;<sup><a href="#param3">3</a></sup>
 *    &lt;param name="liveconnect.DocumentPanelListener" value="documentUpdated" /&gt;<sup><a href="#param4">4</a></sup>
 *    <i>&lt;!-- following parameters improve the experience under Java 1.6u10 --&gt;</i>
 *    &lt;param name="jnlp_href" value="bfopdf-applet.jnlp" /&gt; <i>&lt;!-- see below --&gt;</i><sup><a href="#param5">5</a></sup>
 *    &lt;param name="draggable" value="true" /&gt;
 *    &lt;param name="image" value="http://bfo.com/external/loading.gif" /&gt;
 *    &lt;param name="boxborder" value="false" /&gt;
 *    &lt;param name="centerimage" value="true" /&gt;
 *    &lt;param name="separate_vm" value="true" /&gt; <i>&lt;!-- optional --&gt;</i>
 *   &lt;/applet&gt;
 *   &lt;form onsubmit="return false"&gt;
 *    &lt;input type="button" onclick="document.pdfapplet.runWidgetAction('PagePrevious')" value="&lt;&lt;"/&gt;
 *    &lt;input type="text" name="pagenumber" size="5"
 *        onchange="document.pdfapplet.getViewer().getActiveDocumentPanel().setPageNumber(this.value-1)"/&gt;
 *    &lt;input type="button" onclick="document.pdfapplet.runWidgetAction('PageNext')" value="&gt;&gt;"/&gt;
 *   &lt;/form&gt;
 *  &lt;/body&gt;
 * &lt;/html&gt;
 * </pre>
 * <p>
 * Things of note here are the initialization parameters to the applet:
 * </p
 * <ol class="extrapadding">
 * <li><a name="param1" />the <code>pdf</code> attribute may be set to load a PDF from the specified URL on startup.
 * Multiple PDFs can be loaded at once by specifying the <code>pdf2</code>, <code>pdf3</code> etc.
 * parameters, and if passwords are needed to open those PDFs they can be specified with
 * the corresponding <code>password</code>, <code>password2</code> etc. parameters. Likewise
 * the <code>filename</code>, <code>filename2</code> etc. parameters can be set to initialize
 * the filename of those PDFs.</li>
 * <li><a name="param2" />the <code>feature.<i>nnn</i></code> may be set to false to turn off an individual
 * feature (with <i>nnn</i> being the name or the full class name of the feature)</li>
 * <li><a name="param3" />the <code>feature.<i>nnn</i>.<i>mmm</i></code> may be specified to set a specific
 * <a href="doc-files/initparams.html">initialization parameter</a> on a feature. The parameters
 * that apply to an individual class are documented in the API docs for that class.</li>
 * <li><a name="param4" />the <code>liveconnect.DocumentPanelListener</code> attribute is set to the name of
 * the JavaScript method that should be called whenever a {@link DocumentPanelEvent} is raised.
 * If specified, this method is called with the event object as a single parameter. This also requires the
 * <code>mayscript</code> attribute on the <code>applet</code> tag as seen here.</li>
 * <li><a name="param5" />
 * the Java 1.6u10 parameters, in particular the "jnlp_href" attribute: This optional parameter refers to a
 * JNLP file which is used to provide more information to the Java plugin, including information that
 * allows the download of Jar files to be optimized. This parameter is ignored by older JVMs, and so is
 * worth including to improve the experience for people using the latest JVM. We provide a sample JNLP file
 * as "bfopdf-applet.jnlp" as part of the distribution. The remaining parameters can be used exactly as
 * above or altered as you see fit. There's more information on the new Java plug-in at
 * <a target="_top" href="https://jdk6.dev.java.net/plugin2/jnlp/">java.net</a>
 * </li>
 * </ol>
 *
 * <p>
 * Controlling the applet can be done via the {@link #runWidgetAction} method for simple widgets
 * or, for more complex operations, the {@link #getViewer} method can be used to access and control
 * the {@link PDFViewer} object.
 * </p><p>
 * Remember that although the applet is contained in a Signed Jar, any access to that
 * applet from unsigned JavaScript will be untrusted by default - which means no access to
 * the local filesystem, no printing and so on (although it is possible to
 * <a href="http://www.mozilla.org/projects/security/components/signed-scripts.html">sign the JavaScript</a>
 * to allow this).
 * </p>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class PDFViewerApplet extends JApplet implements DocumentPanelListener, PropertyManager
{
    private PDFViewer viewer;
    private String listenercallback;
    private ActionListener closeListener;
    private boolean draggedOutOfBrowser = false;

    public void init() {

        try {
            System.getProperty("user.home");
            About.putStaticProperty("Applet", "Trusted");
        } catch (Throwable e) {
            About.putStaticProperty("Applet", "Untrusted from "+getDocumentBase());
        }
        List c = new ArrayList(ViewerFeature.getAllFeatures());
        for (Iterator i = c.iterator();i.hasNext();) {
            ViewerFeature feature = (ViewerFeature)i.next();
            String name = feature.getName();
            boolean include = feature.isEnabledByDefault();
            if ("false".equals(getProperty("feature."+name)) || "false".equals(getProperty("feature."+feature.getClass().getName()))) {
                include = false;
            } else if ("true".equals(getProperty("feature."+name)) || "true".equals(getProperty("feature."+feature.getClass().getName()))) {
                include = true;
            }
            if (!include) {
                i.remove();
            }
        }
        listenercallback = getProperty("liveconnect.DocumentPanelListener");
        String lafname = getProperty("swing.defaultlaf");
        if (lafname != null) {
            try {
                UIManager.setLookAndFeel(lafname);
                SwingUtilities.updateComponentTreeUI(PDFViewerApplet.this);
            } catch (Throwable e) { }
        }

        // Hack for Safari: http://lists.apple.com/archives/java-dev/2008/Mar/msg00092.html
        // Still not perfect but at least buttons look OK
        Object frame = SwingUtilities.getAncestorOfClass(java.awt.Frame.class, this);
        boolean webkitapplet = frame!=null && frame.getClass().getName().equals("sun.plugin.viewer.frame.WebKitEmbeddedFrame");
        if (webkitapplet) {
            Color color = new Color(0xD0D0D0, false);
            UIManager.getDefaults().put("ToolBar.background", color);
            UIManager.getDefaults().put("Button.background", color);
        }
        viewer = new PDFViewer(c);
        if (webkitapplet) {
            viewer.setBackground(UIManager.getColor("Button.background"));
        }

        viewer.setPropertyManager(this);
        PDF.setPropertyManager(this);
        viewer.addDocumentPanelListener(this);
        setContentPane(viewer);
        addComponentListener(new ComponentAdapter() {
            public void componentShown(ComponentEvent event) {
                String pdf = getProperty("pdf");
                if (pdf!=null) {
                    try {
                        String filename = getProperty("filename");
                        loadPDF(pdf, getProperty("password"), filename == null ? null : new File(filename));
                    } catch (IOException e) {
                        Util.displayThrowable(e, PDFViewerApplet.this);
                    }
                }
                for (int i=2;(pdf=getProperty("pdf"+i))!=null;i++) {
                    try {
                        String filename = getProperty("filename"+i);
                        loadPDF(pdf, getProperty("password"+i), filename == null ? null : new File(filename));
                    } catch (IOException e) {
                        Util.displayThrowable(e, PDFViewerApplet.this);
                    }
                }
                PDFViewerApplet.this.removeComponentListener(this);
            }
        });
    }

    public URL getURLProperty(String key) throws java.net.MalformedURLException {
        String property = getProperty(key);
        return property==null ? null : new URL(getDocumentBase(), property);
    }

    public String getProperty(String key) {
        return getParameter(key);
    }

    public String getAppletInfo() {
        return "BFO PDF Viewer "+PDF.VERSION+": http://bfo.com/products/pdf";
    }

    public String[][] getParameterInfo() {
        Collection c = ViewerFeature.getAllFeatures();
        String[][] x = new String[c.size()+3][];
        int pos = 0;
        for (Iterator i = c.iterator();i.hasNext();) {
            String name = ((ViewerFeature)i.next()).getName();
            x[pos++] = new String[] { "feature."+name, "boolean", "Enable/Disable the \""+name+"\" feature" };
        }
        x[pos++] = new String[] { "liveconnect.DocumentPanelListener", "String", "The name of the method to call when a DocumentPanelEvent is raised" };
        x[pos++] = new String[] { "pdf", "String", "The URL of the PDF to load on startup" };
        x[pos++] = new String[] { "pdfn", "String", "Additional URLs of the PDF to load on startup (n>=2)" };
        return x;
    }

    /**
     * <p>
     * Load a PDF from the specified URL. If the URL is relative, it is resolved
     * based on the URL of the page containing the applet. Simply calls
     * {@link #loadPDF(String, String) loadPDF(url, null)}.
     * </p><p>
     * Bear in mind if that any methods, including this one, that are called from JavaScript will
     * run as if the applet was untrusted. Consequently this method cannot be called to load PDF's
     * from the local filesystem.
     * </p>
     *
     * @param url the URL of the PDF to load.
     * @since 2.8.5
     */
    public void loadPDF(String url) throws IOException {
        loadPDF(url, null);
    }

    /**
     * <p>
     * Load a PDF from the specified URL. If the URL is relative, it is resolved
     * based on the URL of the page containing the applet. If the <code>password</code>
     * parameter is not null then that password will be used if required, otherwise
     * any required passwords or private keys will be prompted for - for finer control, the {@link PDFViewer#loadPDF(InputStream, EncryptionHandler[], String, File) PDFViewer.loadPDF()}
     * method can be used.
     * </p><p>
     * Bear in mind if that any methods, including this one, that are called from JavaScript will
     * run as if the applet was untrusted. Consequently this method cannot be called to load PDF's
     * from the local filesystem.
     * </p>
     *
     * @param url the URL of the PDF to load.
     * @param password the password to use to open the PDF, or <code>null</code> to prompt if required.
     * @since 2.8.6
     */
    public void loadPDF(String url, String password) throws IOException {
        loadPDF(url, password, null);
    }

    /**
     * <p>
     * Load a PDF from the specified URL. If the URL is relative, it is resolved
     * based on the URL of the page containing the applet. If the <code>password</code>
     * parameter is not null then that password will be used if required, otherwise
     * any required passwords or private keys will be prompted for - for finer control, the {@link PDFViewer#loadPDF(InputStream, EncryptionHandler[], String, File) PDFViewer.loadPDF()}
     * method can be used.
     * </p><p>
     * Bear in mind if that any methods, including this one, that are called from JavaScript will
     * run as if the applet was untrusted. Consequently this method cannot be called to load PDF's
     * from the local filesystem.
     * </p>
     *
     * @param url the URL of the PDF to load.
     * @param password the password to use to open the PDF, or <code>null</code> to prompt if required.
     * @param savefile the default file to use if the PDF is saved after opening. May be null
     * @since 2.11.22
     */
    public void loadPDF(String url, String password, File savefile) throws IOException {
        PDFImporter importer = (PDFImporter)viewer.getFeature(PDFImporter.class);
        if (importer==null) {
            importer = new PDFImporter();
        }
        if (password!=null) {
            Set set = importer.getEncryptionHandlers();
            set.clear();
            StandardEncryptionHandler handler = new StandardEncryptionHandler();
            handler.setUserPassword(password);
            set.add(handler);
        }
        Importer.ImporterTask task = importer.getImporter(viewer, new URL(getDocumentBase(), url));
        if (savefile != null) {
            task.setFile(savefile);
        }
        task.start(viewer, UIManager.getString("PDFViewer.Loading"));
    }

    /**
     * Return the PDF Viewer embedded in this applet
     * @since 2.8.5
     */
    public PDFViewer getViewer() {
        return viewer;
    }

    /**
     * Run the {@link ViewerWidget#action action()} method on the specified
     * ViewerWidget. Typically this would be called from JavaScript to move
     * between pages or similar widget-based actions.
     * @since 2.8.5
     */
    public void runWidgetAction(String name) {
        ViewerFeature feature = viewer.getFeature(name);
        if (feature instanceof ViewerWidget) {
            DocumentPanel panel = viewer.getActiveDocumentPanel();
            if (panel!=null || !((ViewerWidget)feature).isDocumentRequired()) {
                ((ViewerWidget)feature).action(new ViewerEvent(viewer, panel));
            }
        }
    }

    public void documentUpdated(DocumentPanelEvent event) {
        if (listenercallback != null) {
            try {
                // This method requires the "plugin.jar" supplied with the JRE
                // to be in your ClassPath.
                netscape.javascript.JSObject.getWindow(this).call(listenercallback, new Object[] { event });
            }
            catch (Throwable e) {
                // Opera 9 on OS X
                // if applet has been dragged out of the browser we no longer have javascript access
                // this shouldn't happen, as when we drag the applet out we remove the listener,
                // but it still does occasionally on linux, so it's worth trapping
                e.printStackTrace();
            }
        }
    }

    /**
     * Called by the JRE in 1.6u10 when the Applet is being dragged onto the Desktop
     * @since 2.11.1
     */
    public void appletDragStarted() {
        draggedOutOfBrowser = true;
        // applet is in undecorated JFrame here - decorate frame
        Container container = this.getParent();
        while (container != null) {
            if (container instanceof Frame) {
                Frame frame = (Frame)container;
                frame.setTitle("BFO PDF Viewer");
                frame.setResizable(true);
                frame.setUndecorated(false);
                ImageIcon BFO16 = new ImageIcon(PDFViewer.class.getResource("resources/BFO16.png"));
                frame.setIconImage(BFO16.getImage());

                frame.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        if (closeListener != null) {
                            closeListener.actionPerformed(null);
                        }
                    }
                });
                return;
            }
            container = container.getParent();
        }
        viewer.removeDocumentPanelListener(this);
    }

    /**
     * Called by the JRE in 1.6u10 after the Applet has been dragged onto the Desktop
     * @since 2.11.1
     */
    public void appletDragFinished() {
    }

    /**
     * Called by the JRE in 1.6u10, this method signifies that we're drawing
     * our own close button.
     * @param listener the listener to be called when the window is closed
     * @since 2.11.1
     */
    public void setAppletCloseListener(ActionListener listener) {
        closeListener = listener;
    }

    /**
     * Called by the JRE in 1.6u10 when the applet has been closed as a standalone
     * application, indicating the applet has been returned to its location inside
     * the browser.
     * @since 2.11.1
     */
    public void appletRestored() {
        closeListener = null;
        draggedOutOfBrowser = false;
    }

    /**
     * Return flag indicating if we're running outside the browser or not
     * @since 2.11.1
     */
    public boolean isDraggedFromBrowser() {
        return draggedOutOfBrowser;
    }

}
