// $Id: SinglePageDocumentViewport.java 16047 2012-07-16 09:29:03Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.*;
import java.beans.*;
import java.awt.event.*;
import java.awt.geom.Rectangle2D;
import java.util.*;
import javax.swing.*;
import org.faceless.pdf2.*;

/**
 * A type of {@link DocumentViewport} that displays a single page. The Page can be zoomed
 * in or out, and if it grows beyond the size of this viewport scrollbars will be displayed.
 * Zoom levels are translated to DPI (as required by the PagePainter) using the
 * {@link Toolkit#getScreenResolution} method.
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @see DocumentPanel
 * @since 2.8
 */
public class SinglePageDocumentViewport extends DocumentViewport
{
    private JScrollPane scrollPane;
    private View view;
    private float zoom;

    private Listener listener; // event handler
    private RenderingHints hints;
    private ArrayList listeners;
    private ArrayList ilisteners;

    private PDF pdf; // the current PDF document
    private PDFPage page; // the current page
    private PDFParser parser; // parser for use by PagePanel
    private boolean pageWasChanged;

    /**
     * Create a new SinglePageDocumentViewport
     */
    public SinglePageDocumentViewport() {
        view = this.new View();
        setRenderingHints(getDefaultRenderingHints());
        setLayout(new BorderLayout());
        scrollPane = new JScrollPane(view) {
            protected JViewport createViewport() {
                return new JViewport() {
                    public void scrollRectToVisible(Rectangle r) {
                        // NB eat this event to prevent caret from scrolling viewport
                    }
                };
            }
        };
        add(scrollPane, BorderLayout.CENTER);

        // event management
        listener = this.new Listener();
        addComponentListener(listener);
        scrollPane.getHorizontalScrollBar().addAdjustmentListener(listener);
        scrollPane.getVerticalScrollBar().addAdjustmentListener(listener);

        listeners = new ArrayList();
        ilisteners = new ArrayList();

        Util.fixScrollPaneKeyBindings(scrollPane);
    }

    public void setDocumentPanel(DocumentPanel panel) {
        super.setDocumentPanel(panel);
        if (panel!=null) {
            if (this.pdf!=panel.getPDF()) {
                this.page = null;
                this.pdf = panel.getPDF();
                this.parser = panel.getParser();
            }
            scrollPane.getHorizontalScrollBar().setUnitIncrement(mouseWheelUnit);
            scrollPane.getVerticalScrollBar().setUnitIncrement(mouseWheelUnit);

        } else {
            this.pdf = null;
            this.page = null;
            this.parser = null;
            view.dispose();
        }
        if (view.pagePanel != null) {
            view.pagePanel.setParser(parser);
            view.pagePanel.requestFocusInWindow();
        }
    }

    public boolean isDraggable() {
        return view.isDraggable();
    }

    public void setDraggable(boolean draggable) {
        view.setDraggable(draggable);
    }

    public void setPage(PDFPage page, double x, double y, double zoom) {
        if (page.getPDF()!=getDocumentPanel().getPDF()) {
            throw new IllegalArgumentException("Page is from wrong PDF");
        }
        setZoom((float) zoom);
        Dimension pageSize = PagePanel.getFullPageView(page).getBounds().getSize();
        ensureVisible(page, x, ((double) pageSize.height) - y, false);
    }

    public void ensureVisible(PDFPage page, double x, double y) {
        ensureVisible(page, x, y, true);
    }

    /**
     * @param x x coordinate in PDF units
     * @param y y coordinate in PDF units
     * @param centre if true, centre this location in the viewport,
     * otherwise cause it to be displayed at the top left of the viewport.
     */
    private void ensureVisible(PDFPage page, double x, double y, boolean centre) {
        if (view.pagePanel != null) {
            view.forceLayout = true;
            Rectangle pageRect = view.pagePanel.getBounds();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            int xoffset = 0, yoffset = 0;
            if (x==x) {
                xoffset = pointsToPixels((float) x);
            }
            if (y==y) {
                yoffset = pointsToPixels((float) y);
                // convert to AWT axis space
                yoffset = pageRect.height - yoffset;
            }
            xoffset = pageRect.x + xoffset;
            yoffset = pageRect.y + yoffset;
            if (centre) {
                xoffset -= hsb.getVisibleAmount() / 2;
                yoffset -= vsb.getVisibleAmount() / 2;
            }
            hsb.setValue(xoffset - View.VIEW_MARGIN);
            vsb.setValue(yoffset - View.VIEW_MARGIN);
        }
        this.page = page;
        view.setPage(page);
        if (page != getPage()) {
            redraw();
            pageWasChanged = true;
        }
    }

    public boolean isPageVisible(PDFPage page, double x, double y) {
        if (view.pagePanel == null) {
            return false;
        }
        Rectangle pageRect = view.pagePanel.getBounds();
        Rectangle target = (Rectangle) pageRect.clone();
        if (page==getPage()) {
            if (x==x && y==y) {
                float dpi = (float)getZoom() * Util.getScreenResolution(this);
                int xoffset = Math.round((float) x * dpi / 72);
                target.x += xoffset;
                target.width = 8;
                int yoffset = Math.round((float) y * dpi / 72);
                // convert to PDF axis space
                yoffset = pageRect.height - yoffset;
                target.y += yoffset;
                target.height = 8;
                return getClip().intersects(target);
            } else {
                return true;
            }
        } else {
            return false;
        }
    }

    public JComponent getView() {
        return (JComponent)scrollPane.getViewport().getView();
    }

    public float getZoom() {
        if (zoom==0 && page!=null) {
            zoom = getTargetZoom(getZoomMode(), page);
        }
        return zoom;
    }

    private void redraw() {
        // Relayout and repaint
        view.invalidate();
        invalidate();
        validate();
        repaint();
    }

    public void setZoom(float zoom) {
        if (zoom==zoom && zoom > 0.01 && zoom != this.zoom) {
            // Store where we're currently pointing
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            double hv = (double) hsb.getValue() / (double) hsb.getMaximum();
            double vv = (double) vsb.getValue() / (double) vsb.getMaximum();
            // Set zoom and new DPI
            this.zoom = zoom;
            // Set scrollbar values to point to the same location of the
            // document given new zoom level
            listener.setEnabled(false); // don't send adjustment events
            Dimension ps = view.getPreferredSize();
            hsb.setValue((int) Math.round((double) ps.width * hv));
            vsb.setValue((int) Math.round((double) ps.height * vv));
            listener.setEnabled(true);
            redraw();
        }
    }

    int pointsToPixels(float points) {
        return Math.round(points * (getZoom() * Util.getScreenResolution(this) / 72));
    }

    float pixelsToPoints(int pixels) {
        return (float)pixels / (getZoom() * Util.getScreenResolution(this) / 72);
    }

    public Dimension getViewportSize() {
        return scrollPane.getViewport().getSize();
    }

    public PagePanel getPagePanel() {
        return view.pagePanel;
    }

    public Collection getPagePanels() {
        return (view.pagePanel == null) ? Collections.EMPTY_SET : Collections.singleton(view.pagePanel);
    }

    public PDFPage getRenderingPage() {
        return page;
    }

    private Rectangle getClip() {
        JScrollBar hsb = scrollPane.getHorizontalScrollBar();
        JScrollBar vsb = scrollPane.getVerticalScrollBar();
        return new Rectangle(hsb.getValue(), vsb.getValue(), hsb.getVisibleAmount(), vsb.getVisibleAmount());
    }

    public Adjustable getAdjustable(int position) {
        switch (position) {
            case Adjustable.HORIZONTAL:
                return scrollPane.getHorizontalScrollBar();
            case Adjustable.VERTICAL:
                return scrollPane.getVerticalScrollBar();
            default:
                throw new IllegalArgumentException(Integer.toString(position));
        }
    }

    public void setAdjustableValues(int horizontal, int vertical) {
        listener.setEnabled(false);
        scrollPane.getHorizontalScrollBar().setValue(horizontal);
        scrollPane.getVerticalScrollBar().setValue(vertical);
        listener.setEnabled(true);
        view.invalidate();
        view.validate();
        repaint();
    }

    public void setRenderingHints(RenderingHints hints) {
        this.hints = hints;
        if (view.pagePanel != null) {
            view.pagePanel.setRenderingHints(hints);
        }
    }

    public void addPagePanelListener(PagePanelListener listener) {
        listeners.add(listener);
        if (view.pagePanel != null) {
            view.pagePanel.addPagePanelListener(listener);
        }
    }

    public void removePagePanelListener(PagePanelListener listener) {
        listeners.remove(listener);
        if (view.pagePanel != null) {
            view.pagePanel.removePagePanelListener(listener);
        }
    }

    public void addPagePanelInteractionListener(PagePanelInteractionListener listener) {
        ilisteners.add(listener);
        if (view.pagePanel != null) {
            view.pagePanel.addPagePanelInteractionListener(listener);
        }
    }

    public void removePagePanelInteractionListener(PagePanelInteractionListener listener) {
        ilisteners.remove(listener);
        if (view.pagePanel != null) {
            view.pagePanel.removePagePanelInteractionListener(listener);
        }
    }

    PagePanel createPagePanel() {
        final PagePanel pagePanel = new PagePanel();
        pagePanel.setViewport(this);
        pagePanel.setParser(parser);
        pagePanel.setRenderingHints(hints);
        for (Iterator i = listeners.iterator(); i.hasNext(); ) {
            pagePanel.addPagePanelListener((PagePanelListener) i.next());
        }
        for (Iterator i = ilisteners.iterator(); i.hasNext(); ) {
            pagePanel.addPagePanelInteractionListener((PagePanelInteractionListener) i.next());
        }
        return pagePanel;
    }

    public void setZoomMode(int zoommode) {
        super.setZoomMode(zoommode);
        PDFPage page = getPage();
        if (page!=null) {
            setZoom(getTargetZoom(zoommode, page));
        }
    }

    public float getTargetZoom(int zoommode, PDFPage page) {
        float outzoom = zoom;
        if (page!=null) {
            Dimension pps = PagePanel.getFullPageView(page).getBounds().getSize();
            float screenResolution = Util.getScreenResolution(this);
            pps.width = Math.round(pps.width * screenResolution / 72f);
            pps.height = Math.round(pps.height * screenResolution / 72f);
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            Dimension size = new Dimension(hsb.getVisibleAmount(), vsb.getVisibleAmount());
            switch (zoommode) {
                case ZOOM_FITWIDTH:
                    size.width -= vsb.getWidth();
                    outzoom = ((float) size.width / (float) pps.width);
                    break;
                case ZOOM_FITHEIGHT:
                    size.height -= hsb.getHeight();
                    outzoom = ((float) size.height / (float) pps.height);
                    break;
                case ZOOM_FIT:
                    size.width -= vsb.getWidth();
                    size.height -= hsb.getHeight();
                    float zw = ((float) size.width / (float) pps.width);
                    float zh = ((float) size.height / (float) pps.height);
                    outzoom = (zw < zh) ? zw : zh;
                    break;
                default:
                    outzoom = zoom==0 ? 1 : zoom;
            }
        }
        return outzoom;
    }

    private class View extends JPanel implements PagePanelListener, MouseListener, MouseMotionListener {

        static final int VIEW_MARGIN = 0;

        PDFPage lastPage;
        float lastDPI, lastZoom;
        Rectangle lastClip = new Rectangle(0, 0);
        PagePanel pagePanel;
        boolean forceLayout;

        boolean draggable;
        Point mouseDownEvent;
        Point mouseDownScroll;

        View() {
            setLayout(null);
            setFocusable(false);
            addMouseListener(this);
            addMouseMotionListener(this);
        }

        public Dimension getPreferredSize() {
            Dimension ret = new Dimension(0, 0);
            if (page != null) {
                ret = PagePanel.getFullPageView(page).getBounds().getSize();
                // now convert points to pixels
                ret.width = pointsToPixels((float) ret.width);
                ret.height = pointsToPixels((float) ret.height);
            }
            // and add page margins
            ret.height += (VIEW_MARGIN * 2);
            ret.width += (VIEW_MARGIN * 2);
            return ret;
        }

        void setPage(PDFPage page) {
            if (pagePanel == null || pagePanel.getPage() != page) {
                removeAll();
                if (pagePanel!=null) {
                    pagePanel.setParser(null);
                    pagePanel.removePagePanelListener(view);
                    pagePanel.removeMouseListener(view);
                    pagePanel.removeMouseMotionListener(view);
                }
                pagePanel = createPagePanel();
                pagePanel.addPagePanelListener(view);
                pagePanel.addMouseListener(view);
                pagePanel.addMouseMotionListener(view);
                add(pagePanel);
            }
        }

        public void doLayout() {
            Dimension size = getSize();
            Dimension viewport = getPreferredSize();
            Rectangle clip = new Rectangle();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            clip.x = hsb.getValue();
            clip.y = vsb.getValue();
            clip.width = hsb.getVisibleAmount();
            clip.height = vsb.getVisibleAmount();

            int x = VIEW_MARGIN, y = VIEW_MARGIN;
            // centre pages if entirely visible
            if (viewport.width < size.width) {
                x = (size.width - viewport.width) / 2;
            }
            if (viewport.height < size.height) {
                y = (size.height - viewport.height) / 2;
            }
            if (page != null) {
                Dimension pageSize = PagePanel.getFullPageView(page).getBounds().getSize();
                // Convert points to pixels
                int pw = pointsToPixels(pageSize.width);
                int ph = pointsToPixels(pageSize.height);
                Rectangle pr = new Rectangle(x, y, pw, ph);
                pagePanel.setBounds(pr);
                Rectangle2D fullPageView = PagePanel.getFullPageView(page);
                Rectangle pclip = pr.intersection(clip);
                pclip.x -= pr.x;
                pclip.y -= pr.y;
                float zoom = getZoom();
                float dpi = (float)zoom * Util.getScreenResolution(this);
                if (forceLayout || lastPage != page || Math.abs(lastDPI-dpi) >= 0.1 || Math.abs(lastZoom-zoom) >= 0.1 || (!lastClip.equals(pclip))) {
                    pagePanel.setPage(page, fullPageView, dpi, pclip);
                    lastPage = page;
                    lastDPI = dpi;
                    lastZoom = getZoom();
                    lastClip = pclip;
                    forceLayout = false;
                }
            }
        }

        public void paintComponent(Graphics g) {
            if (zoom!=0) {
                Rectangle clip = g.getClipBounds();
                // Paint background
                g.setColor(getBackground());
                g.fillRect(clip.x, clip.y, clip.width, clip.height);
                // Paint page border
                Color dark = new Color(0x66, 0x66, 0x66);
                Color medium = new Color(0x99, 0x99, 0x99);
                Color light = new Color(0xcc, 0xcc, 0xcc);
                Rectangle r = pagePanel.getBounds();

                int x1 = r.x - 1, y1 = r.y - 1;
                int x2 = r.x + r.width + 1, y2 = r.y + r.height + 1;
                g.setColor(dark);
                g.drawRect(x1, y1, r.width + 2, r.height + 2);

                g.setColor(medium);
                x1++;
                y1 = y2 + 1;
                x2--;
                y2++;
                g.drawLine(x1, y1, x2, y2);

                g.setColor(light);
                x1++;
                y1++;
                x2--;
                y2++;
                g.drawLine(x1, y1, x2, y2);
            }
        }

        public void pageUpdated(PagePanelEvent event) {
            if (event.getType()=="redrawn") {
                PagePanel panel = event.getPagePanel();
                panel.invalidate();
                repaint();
                DocumentPanel docpanel;
                if (pageWasChanged && (docpanel = getDocumentPanel())!=null) {
                    pageWasChanged = false;
                    docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPageChanged(docpanel));
                }
            }
        }

        boolean isDraggable() {
            return draggable;
        }

        void setDraggable(boolean draggable) {
            this.draggable = draggable;
            setCursor(draggable ? CURSOR_GRAB : null);
        }

        public void mouseClicked(MouseEvent event) {}
        public void mouseEntered(MouseEvent event) {}
        public void mouseExited(MouseEvent event) {}
        public void mouseMoved(MouseEvent event) {}

        public void mousePressed(MouseEvent event) {
            if (draggable) {
                Component source = (Component) event.getSource();
                if (source == this) {
                    mouseDownEvent = event.getPoint();
                } else {
                    mouseDownEvent = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                mouseDownScroll = new Point(sbh.getValue(), sbv.getValue());
                setCursor(CURSOR_GRABBING);
            }
        }

        public void mouseReleased(MouseEvent event) {
            if (draggable) {
                mouseDownEvent = null;
                mouseDownScroll = null;
                setCursor(CURSOR_GRAB);
            }
        }

        public void mouseDragged(MouseEvent event) {
            if (draggable) {
                Point point;
                Component source = (Component) event.getSource();
                if (source == this) {
                    point = event.getPoint();
                } else {
                    point = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                int dx = point.x - mouseDownEvent.x;
                int dy = point.y - mouseDownEvent.y;
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                int sbx = sbh.getValue();
                int sby = sbv.getValue();
                setAdjustableValues(mouseDownScroll.x - dx, mouseDownScroll.y - dy);
                int sdx = sbh.getValue() - sbx;
                int sdy = sbv.getValue() - sby;
                mouseDownScroll.x += sdx;
                mouseDownScroll.y += sdy;
            }
        }

        void dispose() {
            if (pagePanel!=null) {
                pagePanel.setParser(null);
                remove(pagePanel);
            }
            pagePanel = null;
            lastPage = null;
        }

    }

    public void propertyChange(PropertyChangeEvent event) {
        String name = event.getPropertyName();
        if ((name.endsWith("Box") || name.equals("orientation")) && event.getSource() instanceof PDFPage) {
            redraw();
        }
    }

    private class Listener implements ComponentListener, AdjustmentListener {

        private boolean enabled = true;

        void setEnabled(boolean enabled) {
            this.enabled = enabled;
        }

        public void componentResized(ComponentEvent e) {
            int zoommode = getZoomMode();
            switch (zoommode) {
              case ZOOM_FITWIDTH:
              case ZOOM_FITHEIGHT:
              case ZOOM_FIT:
                setZoom(getTargetZoom(zoommode, getPage()));
                break;
            }
        }

        public void componentMoved(ComponentEvent e) {
        }

        public void componentShown(ComponentEvent e) {
        }

        public void componentHidden(ComponentEvent e) {
        }

        public void adjustmentValueChanged(AdjustmentEvent e) {
            if (enabled) {
                view.invalidate();
                repaint();
                
                DocumentPanel docpanel = getDocumentPanel();
                if (docpanel != null) {
                    docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPagePositionChanged(docpanel));
                }
            }
        }
    }
}
