// $Id: TextSelection.java 15955 2012-06-28 17:09:29Z mike $

package org.faceless.pdf2.viewer2.feature;

import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import org.faceless.pdf2.*;
import org.faceless.pdf2.viewer2.PagePanel;
import java.util.*;
import java.lang.ref.WeakReference;
import java.util.List;
import java.text.AttributedString;
import org.faceless.util.AttributedStringBuilder;

/**
 * This class is a transparent panel which is laid overtop of the PagePanel. It will
 * highlight any text in the {@link RangeList} returned by {@link #getRangeList}.
 * This class is used by the {@link TextTool} and {@link TextHighlighter} classes
 * to manage text markup, and it's unlikely that customers would need to access this
 * class directly.
 */
public abstract class TextSelection extends JPanel {

    private final Paint color;
    private final Stroke stroke;
    private final float margin;
    private final int type;

    /**
     * A type of highlight which will highlight the selected
     * text using a solid block of (usually translucent) color
     */
    public static final int TYPE_BLOCK = 1;

    /**
     * A type of highlight which will highlight the selected
     * text using a single underline.
     * @since 2.11.4
     */
    public static final int TYPE_UNDERLINE = 2;

    /**
     * A type of highlight which will highlight the selected
     * text using a double underline.
     * @since 2.11.4
     */
    public static final int TYPE_DOUBLEUNDERLINE = 4;

    /**
     * A type of highlight which will outline the selected
     * text.
     * @since 2.11.4
     */
    public static final int TYPE_OUTLINE = 8;

    /**
     * A type of highlight which will strike-out the selected
     * text.
     * @since 2.11.4
     */
    public static final int TYPE_STRIKEOUT = 16;

    /**
     * A type of highlight which will strike-out the selected
     * text with a double line
     * @since 2.11.4
     */
    public static final int TYPE_DOUBLESTRIKEOUT = 32;

    public TextSelection(int type, Paint color, Stroke stroke, float margin) {
        this.color = color;
        this.type = type;
        this.stroke = stroke;
        this.margin = margin;
        setOpaque(false);
    }

    /** 
     * Return true if this panel should be treated as a GlassPane. By default
     * this is the case, in which case the cursor will not be modified just
     * because this panel is visible.
     * @since 2.11.25
     */
    public boolean isGlassPane() {
        return true;
    }

    public boolean contains(int x, int y) {
        // Yuck. http://weblogs.java.net/blog/alexfromsun/archive/2006/09/a_wellbehaved_g.html
        return !isGlassPane() && super.contains(x, y);
    }

    /**
     * Return the RangeList that will be highlighted by this TextTool
     */
    protected abstract RangeList getRangeList();

    public void paintComponent(Graphics gg) {
        super.paintComponent(gg);

        PagePanel pagepanel = (PagePanel)getParent();
        PDFPage page = pagepanel.getPage();
        if (page == null) {
            return;             // Redrawing
        }
        Graphics2D g = (Graphics2D)gg;

        if (getClientProperty("pdf.rect") == null) {
            // To ensure we're always resized to match PagePanel
            putClientProperty("pdf.rect", page.getBox("ViewBox"));
        }
        boolean path = type!=TYPE_BLOCK && type!=TYPE_OUTLINE;
        RangeList rangelist = getRangeList();

        for (int j=0;j<rangelist.size();j++) {
            Range range = (Range)rangelist.get(j);
            if (range.isValid()) {
                PageExtractor.Text first = range.getFirst();
                PageExtractor.Text last = range.getLast();
                AffineTransform tran = g.getTransform();
                g.transform(pagepanel.getPageToScreenTransform());
                int firstpos = range.getFirstPosition();
                int lastpos = range.getLastPosition();
                if (first==last) {
                    drawHighlight(type, color, stroke, g, TextTool.getShape(first, firstpos, lastpos, margin, path));
                } else {
                    boolean inside = false;
                    PageExtractor extractor = first.getPageExtractor();
                    Collection c = extractor.getTextUnordered();
                    for (Iterator i = c.iterator();i.hasNext();) {
                        PageExtractor.Text text = (PageExtractor.Text)i.next();
                        if (text==first) {
                            inside = true;
                            drawHighlight(type, color, stroke, g, TextTool.getShape(text, firstpos, text.getTextLength()-1, margin, path));
                        } else if (text==last) {
                            inside = false;
                            drawHighlight(type, color, stroke, g, TextTool.getShape(text, 0, lastpos, margin, path));
                        } else if (inside) {
                            drawHighlight(type, color, stroke, g, TextTool.getShape(text, margin, path));
                        }
                    }
                }
                g.setTransform(tran);
            }
        }
    }

    //-----------------------------------------------------------------------------------------

    /**
     * Draw a highlight of a certain type onto the specified graphics.
     *
     * @param type the shape type
     * @param color the Color
     * @param g the Graphics Object
     * @param shape for Outline or Block, any shape. For underline types,
     * a GeneralPath of the form (x1,y1, x1,y2, x2,y2, x2,y1)
     */
    public void drawHighlight(int type, Paint color, Stroke stroke, Graphics2D g, Shape shape) {
        if (shape == null) {
            return;
        }
        g.setPaint(color);
        if (type==TYPE_BLOCK) {
            g.fill(shape);
        } else if (type==TYPE_OUTLINE) {
            g.draw(shape);
        } else if (type==TYPE_UNDERLINE || type==TYPE_DOUBLEUNDERLINE || type==TYPE_STRIKEOUT || type==TYPE_DOUBLESTRIKEOUT) {
            Point2D.Float p1 = null, p2 = null;

            // Here we assume path is bl, tl, tr, br
            PathIterator i = ((GeneralPath)shape).getPathIterator(new AffineTransform());
            float[][] f = new float[4][6];
            i.currentSegment(f[0]); i.next();
            i.currentSegment(f[1]); i.next();
            i.currentSegment(f[2]); i.next();
            i.currentSegment(f[3]); i.next();

            float linewidth = (float)Math.sqrt((f[1][0]-f[0][0])*(f[1][0]-f[0][0])+(f[1][1]-f[0][1])*(f[1][1]-f[0][1])) / 20;
            if (type==TYPE_STRIKEOUT || type==TYPE_DOUBLESTRIKEOUT) {
                p1 = new Point2D.Float((f[1][0]+f[0][0])/2, (f[1][1]+f[0][1])/2);
                p2 = new Point2D.Float((f[2][0]+f[3][0])/2, (f[2][1]+f[3][1])/2);
            } else {
                p1 = new Point2D.Float(f[0][0], f[0][1]);
                p2 = new Point2D.Float(f[3][0], f[3][1]);
            }
            if (stroke==null) {
                stroke = new BasicStroke(linewidth);
            }
            g.setStroke(stroke);

            Line2D.Float line = new Line2D.Float(p1, p2);
            g.draw(line);
            if (type==TYPE_DOUBLEUNDERLINE || type==TYPE_DOUBLESTRIKEOUT) {
                if (stroke instanceof BasicStroke) {
                    linewidth = ((BasicStroke)stroke).getLineWidth();
                }
                // Move linewidth*2 pixels perpendicular to vector (p1, p2)
                float vx = (float)(p2.getX()-p1.getX());
                float vy = (float)(p2.getY()-p1.getY());
                double len = linewidth * 2 / Math.sqrt(vx*vx+vy*vy);
                g.draw(AffineTransform.getTranslateInstance(vy*len, -vx*len).createTransformedShape(line));
            }
        }
    }

    /**
     * A Range represents a selected range of {@link org.faceless.pdf2.PageExtractor.Text} items.
     * The range may only cover a single page
     */
    public static abstract class Range implements Comparable {

        /**
         * Return the {@link PDFPage} this range object relates to.
         */
        public PDFPage getPage() {
            return isValid() ? getFirst().getPage() : null;
        }

        /**
         * Return the first {@link org.faceless.pdf2.PageExtractor.Text} object that is selected
         */
        public abstract PageExtractor.Text getFirst();

        /**
         * Return the last {@link org.faceless.pdf2.PageExtractor.Text} object that is selected
         */
        public abstract PageExtractor.Text getLast();

        /**
         * Return offset into the Text item returned by {@link #getFirst} that
         * begins the selection.
         */
        public abstract int getFirstPosition();

        /**
         * Return offset into the Text item returned by {@link #getLast} that
         * ends the selection. Note this is the offset of the last selected character,
         * not the first unselected character.
         */
        public abstract int getLastPosition();

        /**
         * Return the text this Range covers
         */
        public StringBuffer getText() {
            return getFirst().getPageExtractor().getText(getFirst(), getFirstPosition(), getLast(), getLastPosition()+1, true);
        }

        /**
         * Return the text this Range covers
         */
        public AttributedString getStyledText() {
            return getFirst().getPageExtractor().getStyledText(getFirst(), getFirstPosition(), getLast(), getLastPosition()+1, true);
        }

        /**
         * Return true if this Range is valid, false otherwise
         */
        public abstract boolean isValid();

        /**
         * Create a new Range that matches the specified text item
         */
        public static Range createRange(PageExtractor.Text text) {
            return new FixedRange(text);
        }

        /**
         * Create a new Range that covers the range of text from <code>first</code>
         * character <code>firstposition</code. to <code>last</code> character <code>lastposition</code>
         * inclusive.
         */
        public static Range createRange(PageExtractor.Text first, int firstposition, PageExtractor.Text last, int lastposition) {
            return new FixedRange(first, firstposition, last, lastposition);
        }

        public int compareTo(Object o) {
            Range range = (Range)o;
            PDFPage page1 = getPage();
            PDFPage page2 = range.getPage();
            int v = page1==page2 ? 0 : page1.getPageNumber() - page2.getPageNumber();
            if (v==0) {
                v = getFirst().compareTo(range.getFirst());
                if (v==0) {
                    v = getFirstPosition() - range.getFirstPosition();
                }
            }
            return v;
        }
    }

    private static class FixedRange extends Range {
        private WeakReference first, last;
        private int firstposition, lastposition;

        FixedRange(PageExtractor.Text text) {
            this (text, 0, text, text.getTextLength()-1);
        }

        FixedRange(PageExtractor.Text first, int firstposition, PageExtractor.Text last, int lastposition) {
            if (first.getPageExtractor()!=last.getPageExtractor()) {
                throw new IllegalArgumentException("PageExtractor mismatch");
            }
            while (first.getPrimaryText()!=null) {
                firstposition += first.getPrimaryTextOffset();
                first = first.getPrimaryText();
            }
            while (last.getPrimaryText()!=null) {
                lastposition += last.getPrimaryTextOffset();
                last = last.getPrimaryText();
            }
            this.first = new WeakReference(first);
            this.last = new WeakReference(last);
            this.firstposition = firstposition;
            this.lastposition = lastposition;
        }

        public PageExtractor.Text getFirst() {
            return (PageExtractor.Text)first.get();
        }

        public PageExtractor.Text getLast() {
            return (PageExtractor.Text)last.get();
        }

        public int getFirstPosition() {
            return firstposition;
        }

        public int getLastPosition() {
            return lastposition;
        }

        boolean concatenate(FixedRange range) {
            if (getLast()==range.getFirst() && getLastPosition() + 1 <= range.getFirstPosition()) {
                last = range.last;
                lastposition = range.lastposition;
                return true;
            } else {
                return false;
            }
        }

        public boolean isValid() {
            return true;
        }
    }

    /**
     * A <code>RangeList</code> is a read-only list of <code>Range</code> objects.
     * The range may cover more than one page.
     */
    public static class RangeList extends AbstractList {

        private final List list;

        static final RangeList EMPTY_RANGELIST = new RangeList(Collections.EMPTY_LIST);

        /**
         * Create a RangeList containing a single {@link Range}
         */
        public RangeList(Range range) {
            this(Collections.singletonList(range));
        }

        /**
         * Create a RangeList containing a list of {@link Range} objects
         * @param list the list of {@link Range} objects
         */
        public RangeList(List list) {
            if (list.size() > 1) {
                Collections.sort(list);
                Range range = (Range)list.get(0);
                for (int i=1;i<list.size();i++) {
                    Range range2 = (Range)list.get(i);
                    if (range instanceof FixedRange && range2 instanceof FixedRange) {
                        if (((FixedRange)range).concatenate((FixedRange)range2)) {
                            list.remove(i--);
                        } else {
                            range = range2;
                        }
                    }
                }
            }
            this.list = list;
        }

        public Object get(int index) {
            return list.get(index);
        }

        public int size() {
            return list.size();
        }

        /**
         * Return true if this list is not empty and every Range it contains is valid,
         * false otherwise
         */
        public boolean isValid() {
            boolean valid = list.size() > 0;
            for (int i=0;valid && i<list.size();i++) {
                valid &= ((Range)list.get(i)).isValid();
            }
            return valid;
        }

        /**
         * Return the list of Pages included with this Range
         * @since 2.11.8
         */
        public Set getPages() {
            LinkedHashSet set = new LinkedHashSet();
            for (int i=0;i<list.size();i++) {
                Range range = (Range)list.get(i);
                set.add(range.getPage());
            }
            return set;
        }

        /**
         * Get the Corners of the area marked by this highlight. These are
         * in the same order as those returned by {@link org.faceless.pdf2.PageExtractor.Text#getCorners}
         * @param page the PDFPage that we want the corners for.
         */
        public float[] getCorners(PDFPage page) {
            List fcorners = new ArrayList();
            for (int i=0;i<list.size();i++) {
                Range range = (Range)list.get(i);
                if (range.getPage()==page) {
                    PageExtractor.Text first = range.getFirst();
                    PageExtractor.Text last = range.getLast();
                    int firstpos = range.getFirstPosition();
                    int lastpos = range.getLastPosition();
                    if (first==last) {
                        int len = lastpos - firstpos;
                        if (len > 0) {
                            PageExtractor.Text sub = first.getSubText(firstpos, len);
                            float[] q = sub.getCorners();
                            for (int k=0;k<8;k++) {
                                fcorners.add(new Float(q[k]));
                            }
                        }
                    } else {
                        boolean inside = false;
                        PageExtractor extractor = first.getPageExtractor();
                        for (Iterator j = extractor.getTextUnordered().iterator();j.hasNext();) {
                            PageExtractor.Text text = (PageExtractor.Text)j.next();
                            if (text==first) {
                                inside = true;
                                int len = first.getTextLength()-firstpos;
                                if (len > 0) {
                                    PageExtractor.Text sub = first.getSubText(firstpos, len);
                                    float[] q = sub.getCorners();
                                    for (int k=0;k<8;k++) {
                                        fcorners.add(new Float(q[k]));
                                    }
                                }
                            } else if (text==last) {
                                inside = false;
                                PageExtractor.Text sub = last.getSubText(0, lastpos+1);
                                float[] q = sub.getCorners();
                                for (int k=0;k<8;k++) {
                                    fcorners.add(new Float(q[k]));
                                }
                            } else if (inside) {
                                float[] q = text.getCorners();
                                for (int k=0;k<8;k++) {
                                    fcorners.add(new Float(q[k]));
                                }
                            }
                        }
                    }
                }
            }
            float[] corners = new float[fcorners.size()];
            for (int i=0;i<corners.length;i++) {
                corners[i] = ((Float)fcorners.get(i)).floatValue();
            }
            return corners;
        }

        /**
         * Return the text included in this RangeList
         */
        public StringBuffer getText() {
            StringBuffer sb = new StringBuffer();
            PDFPage lastpage = null;
            for (int i=0;i<list.size();i++) {
                Range range = (Range)list.get(i);
                if (sb.length() > 0) {
                    sb.append(range.getPage() == lastpage ? ' ' : '\n');
                }
                sb.append(range.getText());
                lastpage = range.getPage();
            }
            return sb;
        }

        /**
         * Return the text included in this RangeList
         */
        public AttributedString getStyledText() {
            AttributedStringBuilder buf = new AttributedStringBuilder();
            PDFPage lastpage = null;
            for (int i=0;i<list.size();i++) {
                Range range = (Range)list.get(i);
                if (buf.length() > 0) {
                    buf.append(range.getPage() == lastpage ? " " : "\n");
                }
                buf.append(range.getStyledText());
                lastpage = range.getPage();
            }
            return buf.toAttributedString();
        }

        public String toString() {
            return list.toString();
        }

    }

}
